#include "config.h"
#include <string.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include "pins.h"

#include "bc.c"
#include "webmsg.c"
#include "prg_avr.c"


#define LCD_DATA   (D,5)
#define LCD_CLOCK  (D,6)
#define LCD_RCLOCK (D,7)

#define MODESEL1   (D,2)
#define MODESEL2   (D,3)

#define APMODEPIN     MODESEL1
#define ERASEPIN      MODESEL2

volatile uint8_t waitchar;

volatile uint8_t have_line;
volatile uint8_t rxbuf[512];
volatile uint16_t rxhead, rxtail;


uint8_t EEMEM es_ssid_work[32]=
  {
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff };
uint8_t EEMEM es_pwd_work[32]=
  {
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff };
uint8_t EEMEM es_ssid_conf[32]= 
  {// 0      1      2      3      4      5      6      7
    'w',    'i',   'f',   'i',   'p',   'r',   'o',   'g',
      0,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff };
uint8_t EEMEM es_pwd_conf[32]=
  {
    'h',    'u',   'n',   't',   'e',   'r',   '2',     0,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff,
   0xff,   0xff,  0xff,  0xff,  0xff,  0xff,  0xff,  0xff };

uint32_t EEMEM es_baud = 9600;

#include "lcd.c"

void lcd_str(char *str)
{
  uint8_t i;
  for(i=0;str[i] && i<20 && str[i]!=0xff;i++) lcd_char(str[i]);
}

void lcd_line(int row, int clr)
{
  uint8_t i;
  lcd_goto(row, 0);
  if (!clr) return ;
  for(i=0;i<20;i++) lcd_char(' ');
  lcd_goto(row,0);
}

void lcd_dec(uint32_t K)
{
  uint32_t D;
  uint8_t V,f;
  f= 0;
  for(D=1000000;D>0;D/=10)
  {
    for(V=0;K>=D;V++)  K-= D;
    if (V || f || D==1) { lcd_char('0'+V); f= 1; }
  }
}

#include "serial.c"
#include "b64.c"

#define LINSIZ 128
uint8_t linebuf[LINSIZ]; uint8_t linelen;
uint8_t resbuf[LINSIZ]; uint8_t reslen;

uint8_t get_eeprom(uint8_t *field)
{
  uint8_t i;
  eeprom_read_block(linebuf, field, 32);
  for(i=0;i<32;i++) if (linebuf[i]==0xff) { linebuf[i]= 0; break; }
  linebuf[31]= 0;
  if (linebuf[0]==0) return 1;
  return 0;
}


void put_eeprom(uint8_t *field)
{
  uint8_t i;
  for(i=0;i<32;i++) if (linebuf[i]==0) break;
  if (i==32) { linebuf[31]= 0; i= 31; }
  for(i++;i<32;i++) linebuf[i]= 0xff;
  eeprom_update_block(linebuf, field, 32);
}

void set_eeprom_str(uint8_t *field, uint8_t *str)
{
  uint8_t i;
  for(i=0;i<32;i++) if (str[i]==0) break;
  if (i==32) { str[31]= 0; i= 31; }
  for(i++;i<32;i++) str[i]= 0xff;
  eeprom_update_block(str, field, 32);
}

int getline()
{
  uint8_t C;
  linelen= 0;
again:
  while(rxhead==rxtail) _delay_us(10);
  if (linelen==LINSIZ) return 0;
  C= rxbuf[rxhead];
  rxhead= (rxhead+1)%512;
  if (linelen<LINSIZ) linebuf[linelen++]= C;
  if (C=='\n') return 1;
  goto again;
}

uint8_t line_starts(char *pfx)
{
  uint8_t i;
  for(i=0;pfx[i] && i<linelen;i++)
   if (pfx[i]!=linebuf[i]) return 0;
  return i;
}

uint8_t str_starts(char *str, uint8_t L, char *pfx)
{
  uint8_t i;
  for(i=0;pfx[i] && i<L;i++)
   if (str[i]!=pfx[i]) return 0;
  return i;
}

uint8_t check_message(char *msg)
{
  uint8_t k;
  if (!(k=line_starts(msg))) return 0;
  if (k<linelen && linebuf[k]=='\r') return 1;
  return 0;
}

uint8_t is_ok() { return check_message("OK"); }
uint8_t is_error() { return check_message("ERROR"); }

void serial_command(char *k)
{
  usart_send(k, strlen(k));
  usart_send("\r\n", 2);
}

uint8_t simple_response()
{
  while(1)
  {
    getline();
    if (is_ok()) return 0;
    if (is_error()) return 1;
  }
}

uint8_t simple_command(char *k)
{
  serial_command(k);
  return simple_response();
}


uint8_t get_str_param(char *pfx)
{
  uint16_t i,k;
  uint8_t result;

  result= 1;
  while(1)
  {
    getline();
    if (is_ok() || is_error()) return result;
    if ((k=line_starts(pfx)))
    {
      for(i=k+1;i<linelen;i++) if (linebuf[i]=='"') { linelen= i; break; }
      for(i=0;i+k<linelen;i++) resbuf[i]= linebuf[i+k];
      reslen= linelen-k; resbuf[reslen]= 0;
      result= 0;
    }
  }
}

uint8_t get_connected_wifi_ssid()
{
  uint8_t result;
  int16_t i,k;
  serial_command("AT+CWJAP?");
  return get_str_param("+CWJAP:\"");
}

uint8_t disconnect_from_wifi() { return simple_command("AT+CWQAP"); }


uint8_t connect_to_wifi()
{
  uint8_t i;
  if (get_connected_wifi_ssid()) goto notconn;
  if (get_eeprom(es_ssid_work) ||strncmp(linebuf, resbuf, 32)) 
    { disconnect_from_wifi(); goto notconn; }

  lcd_clear();
  lcd_pstr(PSTR("W:"));
  for(i=0;resbuf[i]!=0 && resbuf[i]!=0xff && i<32;i++)
    lcd_char(resbuf[i]);

  return 0;
notconn:
  if (get_eeprom(es_ssid_work)) return 1;
  usart_send_str("AT+CWJAP=\"");
  usart_send_str(linebuf);
  usart_send_chr('"');
  if (!get_eeprom(es_pwd_work))
  {
    usart_send_str(",\"");
    usart_send_str(linebuf);
    usart_send_str("\"");
  }
  usart_send_str("\r\n");
  return simple_response();
}

uint8_t get_ip_address()
{
  serial_command("AT+CIFSR");
  return get_str_param("+CIFSR:STAIP,\"");
}

uint8_t enable_server()
{
  if (simple_command("AT+CIPMUX=1")) return 1;
  if (simple_command("AT+CIPSERVER=1,1011")) return 1;
  return 0;
}

uint8_t tcp_send_begin(uint16_t L)
{
  uint32_t timeout;
  uint8_t k;

  usart_send_start();
  usart_send_str_raw("AT+CIPSEND=0,");
  k= 0;
  if (L/100) { usart_send_pure(L/100+'0'); k= 1; L%= 100; }
  if (k || L/10) { usart_send_pure(L/10+'0'); k=1; L%= 10; }
  usart_send_pure(L+'0');
  waitchar= '>';
  usart_send_str_raw("\r\n");
  while(1)
  {
    getline();
      // got the command echoed back
    if (line_starts("AT+CIPSEND=0,")) break;
    if (is_error()) return 1;
  }
  timeout= 500000; // half a second timeout
  while(waitchar=='>')
    { _delay_us(10); timeout-= 10; if (!timeout) return 1; }
  return 0;
}

void tcp_send_end()
{
  usart_send_stop();
  while(1)
  {
    getline();
    if (line_starts("SEND OK")) return ;
    if (is_error()) return ;
  }
}

void w_respond()
{
  uint16_t len;
  len= b64_encoded_len(reslen) + 2;
  if (tcp_send_begin(len)) return ;
  b64_encode_to_serial(resbuf, reslen);
  usart_send_str_raw("\r\n");
  tcp_send_end();
}

void eeprom_setting(uint8_t *tgt, uint8_t *src, uint16_t L)
{
  uint8_t R;
  R= 0xff;
  if (L==0) goto done;
  if (L>LINSIZ-1) L= LINSIZ-1;
  memcpy(linebuf, src, L);
  linebuf[L]= 0;
  put_eeprom(tgt);
  R= 0;
done:
  resbuf[reslen++]= R;
}

void close_connection()
{
  simple_command("AT+CIPCLOSE=0");
}

static uint32_t decode_u32(uint8_t *buf)
{
  uint32_t V, Z,i;
  uint8_t j;
  V= 0;
  for(i=0,j=0;i<32;i+=8,j++) { Z= buf[j]; Z<<= i; V|= Z; }
  return V;
}

static uint16_t decode_u16(uint8_t *buf)
{
  uint16_t V, Z,i;
  uint8_t j;
  V= 0;
  for(i=0,j=0;i<16;i+=8,j++) { Z= buf[j]; Z<<= i; V|= Z; }
  return V;
}

static uint8_t change_baud(uint32_t rate);

static volatile uint8_t* port_sel
   (uint8_t po, volatile uint8_t *b, volatile uint8_t *c, volatile uint8_t *d)
{
  if (po=='B') return b;
  if (po=='C') return c;
  return d;
}

static void port_op(uint8_t po, uint8_t nu, uint8_t cmd)
{
  uint8_t O, A;
  volatile uint8_t *tgt;
  tgt= &O;
  switch(cmd)
  {
  case 'I': case 'C': O= 0; A= ~(1<<nu); break;
  case 'O': case 'S': O= (1<<nu); A= 0xff; break;
  }
  switch(cmd)
  {
  case 'S': case 'C': tgt= port_sel(po, &PORTB, &PORTC, &PORTD); break;
  case 'I': case 'O': tgt= port_sel(po, &DDRB, &DDRC, &DDRD); break;
  }
  *tgt= ((*tgt) | O) & A;
}

void set_target_power(uint8_t value)
{
  // turning off the relay connects power to the target chip
  // because normally closed terminal is tied to VCC, normally open
  // terminal is open
  if (value) { CLR(RELAY_VCC);} else { SET(RELAY_VCC); }
}

void select_target_power(uint8_t value)
{
  if (value) { SET(RELAY_PSEL); } else { CLR(RELAY_PSEL); }
}

void connect_programming_lines(uint8_t value)
{
  // see the comment in relay_init for more details
  // it should have been if(value) set; but I made
  // a mistake during board construction
  if (value) { CLR(RELAY_PRG1); CLR(RELAY_PRG2); }
        else { SET(RELAY_PRG1); SET(RELAY_PRG2); }
}

void run_w_command(uint8_t cmd, uint8_t *buf, uint16_t L)
{
  uint16_t i;
  uint16_t A;
  uint32_t R;
  uint8_t N;

  char *Z= "WiFiBoard";
  switch(cmd)
  {
  case 'I': 
    for(i=0;Z[i];i++) resbuf[reslen++]= Z[i]; 
    if (get_eeprom(es_ssid_conf)) break;
    resbuf[reslen++]= '/';
    for(i=0;linebuf[i];i++) resbuf[reslen++]= linebuf[i];
    break;

  case 'P': eeprom_setting(es_pwd_conf, buf, L); break;
  case 'p': eeprom_setting(es_pwd_work, buf, L); break;
  case 's': eeprom_setting(es_ssid_work, buf, L); break;
  case 'S': eeprom_setting(es_ssid_conf, buf, L); break;

  case '>':  // write to external eeprom. 2 bytes address + 64 bytes data
    // if (L!=66) goto fail;
    A= decode_u16(buf);

    if (bc_write_page(buf+2, A)) 
    { 
     lcd_clear(); lcd_char('?');
     phex(bc_step);
     _delay_ms(1000);
     goto fail; }
    
    resbuf[reslen++]= 0;
    break;
  case '<': // read from external eeprom
            // a status code +  64 bytes of data is returned
            // if the status code is not zero, there is no data.
    //if (L<2) goto fail;
    A= buf[0] | (buf[1]<<8);
    if (bc_read_page(resbuf+reslen+1, A)) // goto fail;
    { 
     lcd_clear(); lcd_char('?');
     phex(bc_step);
     _delay_ms(1000);
     goto fail; }
    
    reslen[resbuf]= 0;
    reslen+= 65;
    break;
  case '!':
    A= buf[0] | (buf[1]<<8);
    self_program_from_external_eeprom(A, linebuf);
    resbuf[reslen++]= 0;
    break;
  case '?':
    A= buf[0] | (buf[1]<<8);
    for(i=0;i<64;i++)
     resbuf[reslen++]= pgm_read_byte(A++);
    break;
  case 'F':
    resbuf[reslen++]= 0x00;
    w_respond();
    close_connection();
    boot_from_external_eeprom();
    break;

  case 'b':
    R= eeprom_read_dword(&es_baud);
    for(i=0;i<4;i++)
      { resbuf[reslen++]= R; R>>= 8; }
    break;
  case 'B':
    R= decode_u32(buf);
    if (R!=eeprom_read_dword(&es_baud)) resbuf[reslen++]= change_baud(R);
                                   else resbuf[reslen++]= 0x00;
    break;

  case '/': // port operations
    port_op(buf[0], buf[1]-'0', buf[2]);
    resbuf[reslen++]= 0;
    break;
  case 'j': connect_programming_lines(buf[0]-'0'); resbuf[reslen++]= 0; break;
  case 'k': select_target_power(buf[0]-'0'); resbuf[reslen++]= 0; break;
  case 'm': set_target_power(buf[0]-'0'); resbuf[reslen++]= 0; break;
  case 'r': if (buf[0]=='0') { CLR(PMAVR_RESET);  }
                        else { SET(PMAVR_RESET); }
            resbuf[reslen++]= 0; break;
  case 'C':
    resbuf[reslen++]= pmAVR_send_command(buf, buf[4]); 
    break;
  case 'a': 
    if (buf[0]<=4) pmAVR_speed= buf[0];
    resbuf[reslen++]= buf[0];
    break;
  case 'W':
    A= decode_u16(buf);
    buf+= 2;
    pmAVR_load_flash_buffer(A, buf+1, buf[0]);
    resbuf[reslen++]= buf[0];
    break;
  case 'R':
    A= decode_u16(buf);
    N= buf[2];
    buf+= 3;
    resbuf[reslen++]= N;
    pmAVR_read_flash(A, resbuf+reslen, N);
    reslen+= N;
    break;
  case 'E':
    A= decode_u16(buf); N= buf[2]; buf+= 3;
    resbuf[reslen++]= N;
    pmAVR_write_eeprom(A, buf, N);
    break;
  case 'e':
    A= decode_u16(buf); N= buf[2]; buf+= 3;
    resbuf[reslen++]= N;
    pmAVR_read_eeprom(A, resbuf+reslen, N);
    reslen+= N;
    break;
  default: goto fail;
  }
  w_respond();
  return ;
fail:
  resbuf[reslen++]= 0xff; 
  w_respond();
}

void process_w_command(uint8_t *buf, uint16_t L)
{
  uint16_t i;
  if (L<5) return ;
     // copy the command identifier immediately to resbuf
  for(i=0;i<4;i++) resbuf[i]= buf[i];
  reslen= 4;
  L-= 4;
  buf+= 4;
  run_w_command(buf[0], buf+1, L-1);
}

void server_loop()
{
  uint16_t L,i,k;
  while(1)
  {
    getline();
    if (line_starts("0,CLOSED")) return ;
    if (line_starts("0,CONNECT")) break;
  }
  lcd_clear();
  lcd_line(0,0); lcd_pstr(PSTR("CONNECTED"));
  while(1)
  {
    getline();
    if (line_starts("0,CLOSED")) break;
    if (!(k=line_starts("+IPD,0,"))) continue;
    L= 0; 
    while(k<linelen && linebuf[k]!=':') { L= L*10 + linebuf[k] - '0'; k++; }
    if (k==linelen || L==0) continue; // bad line
    k++;
    for(i=k;i<linelen;i++) if (linebuf[i]=='\r' || linebuf[i]=='\n') break;
    if (k+L>i) L= i-k; 
     // get only the first line, we will omit the rest
    L= b64_decode(linebuf+k, L);
    process_w_command(linebuf+k, L);
  }
}

uint8_t hexval(uint8_t Z)
{
  if (Z>='0' && Z<='9') return Z-'0';
  if (Z>='A' && Z<='F') return Z-'A'+10;
  if (Z>='a' && Z<='f') return Z-'a'+10;
  return 0;
}

void decode_url_param(char *buf, uint16_t L)
{
  uint16_t i;
  uint8_t V;
  reslen= 0;
  for(i=0;i<L;i++)
  {
    // we have a size limit for our parameters, you know
    if (reslen==32) break;
    if (buf[i]=='+') { resbuf[reslen++]= ' '; continue; }
    if (buf[i]!='%') { resbuf[reslen++]= buf[i]; continue; }
             // not enough characters... just skip it
    i++;
    if (i>=L-1) break;
    resbuf[reslen++]= (hexval(buf[i]) << 4 ) | hexval(buf[i+1]);
    i++;
  }
  resbuf[reslen]= 0;
}

void handle_web_line(char *buf, int16_t L)
{
  uint16_t k;
  uint8_t tgt;
  while(L && buf[0]==' ') { buf++; L--; }
  if (!L) return ;

  for(k=0;k<L;k++) if (buf[k]==' ') return; else if (buf[k]=='?') break;
again:
  buf+= k+1; L-= k+1;
  if (L<=0) return ;
  tgt= buf[0];

  for(k=0;k<L;k++) if (buf[k]==' ') return; else if (buf[k]=='=') break;
  buf+= k+1; L-= k+1;
  if (L<=0) return ;

  for(k=0;k<L;k++) if (buf[k]==' ' || buf[k]=='&') break;
  
  decode_url_param(buf, k);
  if (tgt=='S') 
  { 
    set_eeprom_str( es_ssid_work, resbuf );
    lcd_line(0, 1); lcd_str(resbuf); 
  }
  else if (tgt=='P') 
  { 
    set_eeprom_str( es_pwd_work, resbuf);
    lcd_line(1,1); lcd_str(resbuf); 
  }

  if (k==L || buf[k]==' ') return ;
  goto again;
}

void send_200_OK()
{
  uint8_t i;
  uint16_t A;
  A= (uint16_t) weblines;
  for(i=0;i<5;i++)
  {
    uint16_t L;
    const char *S;
    S= (const char*) pgm_read_word((void*) A);
    L= strlen_P(S);
    if (tcp_send_begin(L)) return ;
    usart_send_progmem(S);
    tcp_send_end();
    _delay_ms(5);
    A+= 2;
  }
}


void web_loop()
{
  uint16_t L,i,k;
  char *buf;
  while(1)
  {
    getline();
    if (line_starts("0,CLOSED")) return ;
    if (line_starts("0,CONNECT")) break;
  }
  lcd_clear();
  lcd_line(0,0); lcd_pstr(PSTR("WEBCONN"));
  while(1)
  {
    getline();
    if (line_starts("0,CLOSED")) break;
    if (!(k=line_starts("+IPD,0,"))) continue;
    for(;k<linelen;k++) if (linebuf[k]==':') break;
    k++;
    if (k>=linelen) continue;
    buf= linebuf+k;
    L= linelen - k;
    if (!(k=str_starts(buf,L,"GET "))) continue;
    handle_web_line(buf+k, L-k);
    send_200_OK();
  }
}

void lcd_progmem(const char *str)
{
  uint8_t K;
  while(1)
  {
    K= pgm_read_byte(str);
    if (!K) break;
    lcd_char(K);
    str++;
  }
}

static void station_loop()
{
  lcd_line(1,1);
  lcd_pstr(PSTR("STATION"));
  simple_command("AT+CWMODE=1");
  while(1)
  {
     if (connect_to_wifi()) { _delay_ms(10000); continue; }
     get_eeprom(es_ssid_work);
     lcd_line(0, 1); lcd_str("W:"); lcd_str(linebuf);
     if (enable_server()) { _delay_ms(1000); continue; }
     get_ip_address();
     lcd_line(1, 1); lcd_str(resbuf);
     server_loop();
  }
}

void set_eeprom_default_string(uint8_t *dst, const uint8_t *pstr)
{
  uint8_t Z;
  do { Z= pgm_read_byte(pstr++); eeprom_write_byte(dst++, Z); } while(Z);
}

uint8_t get_ap_ip_address()
{
  serial_command("AT+CIPAP?");
  return get_str_param("+CIPAP:ip:\"");
}

uint8_t configure_ap()
{
  uint8_t R;
  R= simple_command("AT+CWMODE=2");
   if (R) {
     lcd_line(0,1); lcd_str("mode chg"); return 1;
   }
again:
  if (get_eeprom(es_ssid_conf) || get_eeprom(es_pwd_conf))
  {
    set_eeprom_default_string(es_ssid_conf, PSTR("wifiprog"));
    set_eeprom_default_string(es_pwd_conf, PSTR("hunter2"));
    _delay_ms(4000);
    goto again;
  }

  usart_send_start();
  usart_send_progmem(PSTR("AT+CWSAP=\""));
  get_eeprom(es_ssid_conf);
  usart_send_str_raw(linebuf);
  usart_send_progmem(PSTR("\",\""));  
  get_eeprom(es_pwd_conf);
  usart_send_str_raw(linebuf);
  usart_send_progmem(PSTR("\",5,3\r\n"));  
  usart_send_stop();

  if (simple_response()) { lcd_line(0,1); lcd_str("cwsap"); return 1; }

  if (get_ap_ip_address()) { lcd_line(0,1); lcd_str("apip"); return 1;  }
  lcd_line(1,1);
  lcd_str(resbuf);  
}

void ap_loop()
{
  lcd_line(1,1);
  lcd_str("AP");
  configure_ap();
  enable_server();
  get_eeprom(es_ssid_conf);
  lcd_line(0,1); lcd_str("A:"); lcd_str(linebuf); 
  while(1)
  {
    web_loop();
  }
}

uint8_t get_line_timeout(uint32_t tt)
{
  uint8_t C;
  linelen= 0;
again:
  while(rxhead==rxtail) { _delay_us(10); tt-= 10; if (tt==0) return 1; }
  C= rxbuf[rxhead];
  rxhead= (rxhead+1)%512;
  if (linelen<LINSIZ) linebuf[linelen++]= C;
  if (C=='\n') return 0;
  goto again;
}

/*********************************************************************
  This function sets the transfer rate for the serial line and sends
  out a test command. If we can read the response, then the rate
  is the correct one and the return value is zero. Otherwise, it's 1.
 *********************************************************************/
uint8_t at_test(uint32_t rate)
{
  uint8_t lc;
  lcd_line(0,1);
  usart_set_rate(rate);
  rxhead= rxtail;
  serial_command("AT");
   // timeout is in microseconds, total 10 ms
  if (get_line_timeout(1000*10)) return 1;

  serial_command("AT");
  for(lc=0;lc<100;lc++)
  {
    if (get_line_timeout(1000*10)) return 1;
    if (is_error() || is_ok()) return 0;
  }
  return 1;
}

void wait_idle(const char* pstr)
{
  lcd_pstr(pstr);
  while(1) _delay_ms(1000); // nothing else will work, just stop here
}

void setup_rate()
{
  const static uint8_t mul[]= { 24, 16, 12, 8, 6, 4, 3, 2, 0 };
  uint8_t i;
  uint32_t R,E;

  E= R= eeprom_read_dword(&es_baud);
  if (R>=9600 && R<=115200 && !at_test(R)) goto success;

  for(i=0;mul[i];i++)
   if (!at_test(R=(uint32_t)mul[i] *4800)) goto success;

  wait_idle(PSTR("ratefail"));
success:
  lcd_char('='); lcd_dec(R); 
  if (E!=R) eeprom_write_dword(&es_baud, R);
  _delay_ms(5000);
}

static void check_erase_pin()
{
  uint8_t i;
  if (TST(ERASEPIN)) return ;
  for(i=0;i<32;i++) linebuf[i]= 0xff;

  eeprom_update_block(linebuf, es_ssid_work, 32);
  eeprom_update_block(linebuf, es_pwd_work, 32);
  eeprom_update_block(linebuf, es_ssid_conf, 32);
  eeprom_update_block(linebuf, es_pwd_conf, 32);
  eeprom_update_block(linebuf, &es_baud, 4);

  lcd_line(1,0);
  wait_idle(PSTR("ERASED.REBOOT"));
}

static uint8_t change_baud(uint32_t rate)
{
  usart_send_start();
  usart_send_progmem(PSTR("AT+CIOBAUD="));
  usart_send_decimal(rate);
  usart_send_progmem(PSTR("\r\n"));
  usart_send_stop();
  if (simple_response()) return 1; 
  usart_set_rate(rate);
  eeprom_write_dword(&es_baud, rate);
  return 0;
}

void relay_init()
{
  OUTPUT(RELAY_VCC); CLR(RELAY_VCC);
  OUTPUT(RELAY_PSEL); CLR(RELAY_PSEL);
  /*****************************************************
               WARNING
    While I was building the board, I connected relay_prg2
    incorrectly and realised my error way too late. I 
    connected the normally-open terminal to the pin headers
    and the other one to the lines coming from the programmer.
    Now, this means that in order for the target circuit to
    function on its own, the programmer must be connected 
    and these relays must be driven ON. If I make another 
    one, I won't make the same mistake.
   *****************************************************/
  OUTPUT(RELAY_PRG1); SET(RELAY_PRG1);
  OUTPUT(RELAY_PRG2); SET(RELAY_PRG2);
}


void main()
{
  bc_init();
  relay_init();
  pmAVR_init();
  _delay_ms(2500);
  lcd_init();
  lcd_clear(); 
  lcd_str("READY");
  
  _delay_ms(1000);
  INPUT(APMODEPIN);
  INPUT(ERASEPIN);

  check_erase_pin();

  sei();
  // usart_set_rate(9600);
  setup_rate();

  close_connection(); // signal to others that we have reset
  
  if (TST(APMODEPIN)) station_loop();
                 else ap_loop();
}
