
static uint8_t pmAVR_speed=3;

void pmAVR_init()
{
  OUTPUT(PMAVR_MOSI);      CLR(PMAVR_MOSI);
  OUTPUT(PMAVR_SCK);       CLR(PMAVR_SCK);
  INPUT(PMAVR_MISO);       CLR(PMAVR_MISO);
  OUTPUT(PMAVR_RESET);     CLR(PMAVR_RESET);
}

static void pmAVR_delay()
{
  switch(pmAVR_speed)
  {
  case 0: _delay_us(5); break;
  case 1: _delay_us(10); break;
  case 2: _delay_us(25); break;
  case 3: _delay_us(100); break;
  default: _delay_us(500); break;
  }
}

uint8_t pmAVR_send_byte(uint8_t V)
{
  uint8_t R;
  int i;
  R= 0;
  for(i=0;i<8;i++)
  {
    if (V&(1<<7)) { SET(PMAVR_MOSI); }
             else { CLR(PMAVR_MOSI); }
    pmAVR_delay();
    SET(PMAVR_SCK);
    pmAVR_delay();

    R<<= 1;
    if (TST(PMAVR_MISO)) R|= 1;
    CLR(PMAVR_SCK);
    V<<= 1;
  }
  return R;
}

uint8_t pmAVR_send_command(uint8_t *cmd, uint8_t retb)
{
  uint8_t i;
  uint8_t R,K;
  R= 0;
  for(i=0;i<4;i++)
  {
    K= pmAVR_send_byte(cmd[i]);
    if (i==retb) R= K;
  }
  return R;
}

void pmAVR_load_flash_buffer(uint16_t address, uint8_t *data, uint16_t len)
{
  int i;
  uint8_t hilow;

  if (address&1) hilow= 0x08;
            else hilow= 0x00;
 
  for(i=0;i<len;i++)
  {
    pmAVR_send_byte( 0x40  | hilow );
    pmAVR_send_byte( address >> 9 );
    pmAVR_send_byte( address >> 1 );
    pmAVR_send_byte( data[i] );
    address++;
    hilow ^= 0x08;
  }
}

void pmAVR_read_flash(uint16_t address,uint8_t *data,uint16_t len)
{
  uint16_t i;
  uint8_t hilow;

  if (address&1) hilow= 0x08;
            else hilow= 0x00;
 
  for(i=0;i<len;i++)
  {
    pmAVR_send_byte( 0x20 | hilow );
    pmAVR_send_byte( address >> 9 );
    pmAVR_send_byte( address >> 1 );
    data[i]= pmAVR_send_byte( 0x00 );
    address++;
    hilow ^= 0x08;
  }
}

void pmAVR_read_eeprom(uint16_t address, uint8_t *data, uint16_t len)
{
  uint16_t i;
  for(i=0;i<len;i++)
  {
    pmAVR_send_byte( 0xA0 );
    pmAVR_send_byte(address>>8);
    pmAVR_send_byte(address);
    data[i]= pmAVR_send_byte( 0x00 );
    address++; 
  }
}

uint8_t pmAVR_poll_rdy(uint8_t count)
{
  // this is so stupid, they call it a RDY/_BUSY result, but
  // the LSB is SET while the MCU is busy. so, the 
  // device is ready for more when this bit is cleared
  while(count--)
  {
    pmAVR_send_byte( 0xF0 );
    pmAVR_send_byte( 0x00 );
    pmAVR_send_byte( 0x00 );
    if ((pmAVR_send_byte( 0x00 )&1)==0) return 0;
  }
  return 1;
}

void pmAVR_write_eeprom(uint16_t address, uint8_t *data, uint16_t len)
{
  uint16_t i;
  for(i=0;i<len;i++)
  {
    pmAVR_send_byte( 0xC0 );
    pmAVR_send_byte(address>>8);
    pmAVR_send_byte(address);
    pmAVR_send_byte( data[i] );
    address++; 
    if (pmAVR_poll_rdy(100)) _delay_ms(5); 
  }
}

