#include <stdlib.h>
#include <stdarg.h>
#include "config.h"
#include "chip.h"

#define KB  1024

static char *m328_exfuses[8]=
{
  " 7 // Unused ",
  " 6 // Unused ",
  " 5 // Unused ",
  " 4 // Unused ",
  " 3 // Unused ",
  " 2 // BODLEVEL2 ",
  " 1 // BODLEVEL1 ",
  " 0 // BODLEVEL0 "
};
static char *m328_hifuses[8]=
{
  " 7 // RSTDISBL",
  " 6 // DWEN",
  " 5 // SPIEN",
  " 4 // WDTON",
  " 3 // EESAVE",
  " 2 // BOOTSZ1",
  " 1 // BOOTSZ0",
  " 0 // BOOTRST"
};
static char *m328_lofuses[8]=
{
  " 7 // CKDIV8",
  " 6 // CKOUT",
  " 5 // SUT1",
  " 4 // SUT0",
  " 3 // CKSEL3",
  " 2 // CKSEL2",
  " 1 // CKSEL1",
  " 0 // CKSEL0"
};
static char *m8_hifuses[8]=
{
  " 7 // RSTDISBL",
  " 6 // WDTON",
  " 5 // SPIEN",
  " 4 // CKOPT",
  " 3 // EESAVE",
  " 2 // BOOTSZ1",
  " 1 // BOOTSZ0",
  " 0 // BOOTRST"
};
static char *m8_lofuses[8]=
{
  " 7 // BODLEVEL",
  " 6 // BODEN",
  " 5 // SUT1",
  " 4 // SUT0",
  " 3 // CKSEL3",
  " 2 // CKSEL2",
  " 1 // CKSEL1",
  " 0 // CKSEL0"
};

chip_t chips[]=
{
  { "Atmega328P", { 0x1e, 0x95, 0x0f }, 32*KB, 128, 1*KB,
       "EHL", m328_exfuses, m328_hifuses, m328_lofuses },
  { "Atmega328",  { 0x1e, 0x95, 0x14 }, 32*KB, 128, 1*KB,
       "EHL", m328_exfuses, m328_hifuses, m328_lofuses },
  { "Atmega8A",   { 0x1e, 0x93, 0x07 },  8*KB,  64,  512,
       "HL",  NULL,         m8_hifuses, m8_lofuses },
  { NULL }
};

chip_t *get_chip()
{
  int i;
  if (!config.G_device || !config.device) return NULL;
  for(i=0;chips[i].name;i++)
   if (!strcmp(chips[i].name,config.device)) return chips+i;
  return NULL;
}

int chip_is(chip_t *C, ...)
{
  va_list ap;
  int count;
  char *name;
  va_start(ap,C);
  for(count=0;count<100;count++)
  {
    name= va_arg(ap, char*);
    if (!name) return 0;
    if (!strcmp(name, C->name)) return 1;
  }
  return 0;
}

