#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stddef.h>
#include <time.h>
#include <stdint.h>
#include <stdio.h>
#include "gfx.h"
#include "ticktock.h"
#include "worker.h"
#include "config.h"
#include "confile.h"
#include "chip.h"
#include "sck.h"

int device_socket=-1;
static int tick_fd;

static int main_fd;
static void (*main_comp_func)();
uint32_t main_i1, main_i2;
void *main_p1, *main_p2;

cfg_top_t config;

sck_value_t sck_values[]=
 {
   { 0, "100Khz" },
   { 1, "50Khz" },
   { 2, "20Khz" },
   { 3, "5Khz" },
   { 4, "1Khz" },
   { 5, NULL }
 };

static char *flash_file_types[]= { "Intel Hex File (*.hex)", NULL };
static char *eeprom_file_types[]= { "Intel Hex File (*.eep)", NULL };

static void show_about()
{
  gfx_logview_clear(Wlog);
  loprintf("WiFi Programmer - 1.0 built on 27.10.2017\n"
           "Copyright 2017 Sukru Cinar\n");
}

static int bad_str(char *k)
{
  int i;
  if (!k || !k[0]) return 1;
  for(i=0;k[i];i++)
   if (k[i]<' ' || k[i]>=127 || k[i]=='"') return 1;
  return 0;
}

static int tick_event(gfx_event_t *ev)
{
  char c;
  if (ev->type!=GFXEV_READY_READ) return 0;
  switch(ev->key)
  {
  case 20: ticktock_ack(); worker_time_elapsed(100); break;
  case 30: worker_read_data(); break; 
  case 40: read(main_fd,&c,1);
           if (main_comp_func) {
             void (*f)(); f= main_comp_func;
             main_comp_func=NULL; (*f)(); }
           break;
  default: return 0;
  }
  return 1;
}


uint8_t fuse_edit(uint8_t orig, char *title, char **names)
{
  static int bit_widgets[8]= { 
    WfbBIT0, WfbBIT1, WfbBIT2, WfbBIT3,
    WfbBIT4, WfbBIT5, WfbBIT6, WfbBIT7 };
  static int lab_widgets[8]= {
    WLfbBIT0, WLfbBIT1, WLfbBIT2, WLfbBIT3,
    WLfbBIT4, WLfbBIT5, WLfbBIT6, WLfbBIT7,
  };
  gfx_event_t event;
  int bit;
  char buf[4];
  gfx_FUSEBITS();

  int i;
  uint8_t value;

  for(i=0;i<8;i++)
  {
    // WARNING: names are listed in reverse order because
    // of the way it appears in the data sheet. so, I need
    // to use 7-i instead of just i for names[]
    gfx_set_text(lab_widgets[i], names[7-i]);
    if (orig&(1<<i)) buf[0]= '1'; else buf[0]= '0';
    buf[1]= 0;
    gfx_set_text(bit_widgets[i], buf);
  }
  value= orig;
  gfx_set_text(WLfusebitsIDENT,title);

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case WfusebitsCANCEL:
  case GFXEV_QUIT: value= orig; goto done;
  case WfusebitsOK: goto done;
  case WTfbBIT0: bit= 0; goto change;
  case WTfbBIT1: bit= 1; goto change;
  case WTfbBIT2: bit= 2; goto change;
  case WTfbBIT3: bit= 3; goto change;
  case WTfbBIT4: bit= 4; goto change;
  case WTfbBIT5: bit= 5; goto change;
  case WTfbBIT6: bit= 6; goto change;
  case WTfbBIT7: bit= 7; goto change;
  default:
    goto again;
  }
done:
  gfx_leave_screen();
  return value;
change:
  value^= (1<<bit);
  if (value&(1<<bit)) buf[0]= '1'; else buf[0]= '0';
  buf[1]= 0;
  gfx_set_text(bit_widgets[bit], buf);
  goto again;
}


void ap_settings()
{
  char *s;
  char *ssid, *pwd;
  gfx_event_t event;
  gfx_AP_SETTING();
  
again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT: 
  case WapCANCEL: 
    break;
  case WapCHANGE:
    ssid= gfx_get_text(WapSSID);
    pwd= gfx_get_text(WapPWD);
    if (bad_str(ssid)) { gfx_logview_printf(Wlog,"Bad SSID\n"); goto again; }
    if (bad_str(pwd)) { gfx_logview_printf(Wlog,"Bad PWD\n"); goto again; }
    worker_run('A', ssid, pwd, 0, 0);
    break;
  default:
    goto again;
  }
  gfx_leave_screen();
}

void got_baud_rate()
{
  gfx_event_t event;
  int i,K, B;
  uint32_t rates[]= { 9600, 14400, 19200, 28800, 
                      38400, 57600, 76800, 115200, 0 };
  char buf[20];

  B= main_i1;
  if (B<0) return ;

  gfx_BAUD_SETTING();

  for(i=0;rates[i];i++)
  {
    sprintf(buf,"%d", rates[i]);
    gfx_insert(Wbaud, -1, buf, NULL, NULL);
    if (B==rates[i]) gfx_set_selected(Wbaud, i); 
  }

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT: 
  case WbaudrateCANCEL: 
    K= -1;
    goto done;
  default:
    goto again;
  case WbaudrateOK: 
    K= gfx_get_selected(Wbaud);
    goto done;
  }
done:
  gfx_leave_screen();
  if (K<0 || K>=i) return ;
  gfx_logview_printf(Wlog, "Changing baud rate to %d\n",rates[K]);
  worker_run('B', NULL, NULL, rates[K], 0);
}

void pin_test_screen()
{
  gfx_event_t event;
  char* s;
  int op;

  gfx_PINTEST();

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case WpintestCLOSE:
  case GFXEV_QUIT: 
    goto done;
  case WpintestINPUT:  op= 'I'; goto run;
  case WpintestOUTPUT: op= 'O'; goto run;
  case WpintestSET:    op= 'S'; goto run;
  case WpintestCLR:    op= 'C'; goto run;
  case WpintestDEF: 
    goto again;
  default:
    goto again;
  }
done:
  gfx_leave_screen();
  return ;
run:
  s= gfx_get_text(WpintestPIN);
  if (!s || !s[0] || !s[1]) goto again;
  if (s[0]<'B' || s[0]>'D') goto again;
  if (s[1]<'0' || s[1]>'7') goto again;
  worker_run('/', s, NULL, op, 0);
  goto again;  
}

void firmware_update_screen(char *fn)
{
  char *baddr;
  uint32_t addr;
  gfx_event_t event;
  gfx_FIRMUPD();

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT: 
  case WfirmCANCEL: 
    goto done;
  case WfirmOK:
    baddr= gfx_get_text(WfirmADDR);
    if (!baddr || !baddr[0]) goto again;
    if (gfx_ask("Confirm Firmware Update", "!OK!Cancel", "Are you sure you got this right?\nFile: %s\nBoot sector start: %s", fn, baddr)==0) goto doit;
  default:
    goto again;
  }
done:
  gfx_leave_screen();
  return ;
doit:
  gfx_leave_screen();
  worker_run('F', fn, NULL, strtol(baddr, NULL,0), 0);
  return ;
}

void read_config()
{
  char *e;
  char *fn;
  fn= get_config_file_name("wifiprog","config.txt");
  e= cfg_parse_top(fn,&config);
  if (e) { config.G_verbose_ident= 1;
           config.verbose_ident= 1;
           config.G_device= 1;
           config.device= "Atmega328P";
          loprintf("error reading config file %s:\n   %s\n",fn,e);
          }
     else loprintf("configuration read from %s\n", fn);
}

void write_config()
{
  char *e;
  char *fn;
  char *K;
  fn= get_config_file_name("wifiprog","config.txt");
  config.G_eeprom_file= 0;
  config.G_flash_file= 0;
  K= gfx_get_text(Wflashfile);
  if (K && K[0]) { config.G_flash_file= 1; config.flash_file= K; }
  K= gfx_get_text(Weepromfile);
  if (K && K[0]) { config.G_eeprom_file= 1; config.eeprom_file= K; }
  e= cfg_write_top(fn,&config);
  if (e) loprintf("error writing config file %s:\n   %s\n",fn,e);
     else loprintf("configuration written to %s\n", fn);
}

void maintenance_screen()
{
  gfx_event_t event;
  static char *file_types[]= { "Intel Hex File (*.hex)", NULL };
  char *ffile;

  gfx_MAINT();

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT: 
  case WmaintCLOSE: 
    goto done;
  case WmaintAP: 
    gfx_leave_screen(); 
    ap_settings(); 
    return ;
  case WmaintBAUD:  
    gfx_leave_screen(); 
    main_comp_func= got_baud_rate;
    worker_run('b', 0, 0, 0, 0);
    return ;
  case WmaintPIN:
    gfx_leave_screen();
    pin_test_screen();
    return ;
  case WmaintFIRMWARE:
    ffile= gfx_select_file(0,NULL,file_types);
    if (!ffile) { loprintf("Firmware update cancelled\n"); goto again; }
    gfx_leave_screen();
    firmware_update_screen(ffile); return ;
    break;
  default:
    goto again;
  }
done:
  gfx_leave_screen();
}


char *known_devices[]= { "Atmega328P", "Atmega8", NULL };

void settings_screen()
{
  int i,R;
  gfx_event_t event;
  gfx_SETTINGS();

  if (config.verbose_ident)
   gfx_set_state(WsettingsVERBOSEIDENT,1);

  for(i=0;chips[i].name;i++)
    gfx_insert(WsettingsDEVICE, -1,chips[i].name,NULL,NULL);

  if (config.G_device && config.device)
    for(i=0;known_devices[i];i++)
      if (!strcmp(config.device,chips[i].name))
       { gfx_set_selected(WsettingsDEVICE,i); break; }

  for(i=0;sck_values[i].freq;i++)
    gfx_insert(WsettingsSCKFREQ, -1, sck_values[i].freq, NULL, NULL);

  if (config.G_sck)
    for(i=0;sck_values[i].freq;i++)
       if (sck_values[i].value==config.sck)
        { gfx_set_selected(WsettingsSCKFREQ, i); break; }

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case GFXEV_QUIT: 
  case WsettingsCANCEL: 
    goto done;
  case WsettingsOK:
    goto record;
  default:
    goto again;
  }
done:
  gfx_leave_screen();
  return ;
record:
  config.G_verbose_ident= 1;
  config.verbose_ident= gfx_get_state(WsettingsVERBOSEIDENT);
  R= gfx_get_selected(WsettingsDEVICE);
  if (R>=0) { config.G_device= 1; config.device= chips[R].name; }
  else { config.G_device= 0; config.device= ""; }

  R= gfx_get_selected(WsettingsSCKFREQ);
  if (R>=0) { config.G_sck= 1; config.sck= sck_values[R].value; }
  else { config.G_sck= 0; }
  goto done;
}

static void power_on() 
{
  loprintf("power on\n"); 
  worker_run('G', "m1",NULL,0,0); 
}
static void power_off(int sigcom) 
{
  loprintf("power off\n"); 
  worker_run('G', "m0",NULL,sigcom,0); 
}

static void put_prg(int which)
{
  char *ff,*ef;

  ff= gfx_get_text(Wflashfile); if (!ff || !ff[0]) ff= NULL;
  ef= gfx_get_text(Weepromfile); if (!ef || !ef[0]) ef= NULL;

  switch(which)
  {
  case 1:
    ef= NULL;
    if (!ff) { loprintf("error: no flash file selected\n"); return ; }
    break;
  case 2:
    ff= NULL;
    if (!ef) { loprintf("error: no eeprom file selected\n"); return ; }
    break;
  case 3:
    if (!ff && !ef) { loprintf("error no file selected\n"); return ; }
    break;
  }
  worker_run('P', ff, ef, 0, 0);
}

static void continue_recycle()
{
  msleep(500);
  power_on();
}

static void set_label_hex(int widget, uint8_t V)
{
  char buf[10];
  sprintf(buf,"%02x",V);
  gfx_set_text(widget,buf);
}

static void show_fuses()
{
  int i,R;
  gfx_event_t event;
  chip_t *C;
  uint8_t ex, hi, lo;
  uint32_t o_fuses;
  uint32_t fuses;

  o_fuses= main_i1;
  ex= main_i1 >> 16;
  hi= main_i1 >> 8;
  lo= main_i1;

  C= get_chip();
  if (!C) { loprintf("please select a valid device\n"); return ; }

  gfx_FUSES();

putall:
  if (!strchr(C->fuses,'E'))
  {
    gfx_set_text(WLfusesEX,"-- Not Applicable --");
    gfx_set_text(WfusesEX, "ZZ");
  }
  else
  {
    set_label_hex(WfusesEX, ex);
  }
  set_label_hex(WfusesHI, hi);
  set_label_hex(WfusesLO, lo);

again:
  if (!gfx_event(NULL,&event)) goto again;
  if (tick_event(&event)) goto again;
  switch(event.type)
  {
  case WfusesCANCEL:
  case GFXEV_QUIT: 
    goto done;
  case WfusesOK: goto write_fuses;
  case WCfusesEX: 
    if (!strchr(C->fuses,'E')) goto again;
    ex= fuse_edit(ex, "Extended Fuse Byte", C->ex_fuse_names);
    goto putall;
  case WCfusesHI: 
    if (!strchr(C->fuses,'H')) goto again;
    hi= fuse_edit(hi, "High Fuse Byte", C->hi_fuse_names);
    goto putall;
  case WCfusesLO: 
    if (!strchr(C->fuses,'L')) goto again;
    lo= fuse_edit(lo, "Low Fuse Byte", C->lo_fuse_names);
    goto putall;
  default:
    goto again;
  }
done:
  gfx_leave_screen();
  return ;
write_fuses:
  fuses= (ex<<16) | (hi<<8) | lo;
  worker_run('U', NULL, NULL, fuses, o_fuses);
  goto done;
}

static void save_eeprom()
{
  char *fn;
  fn= gfx_select_file(1, NULL, eeprom_file_types);
  if (!fn) return ;
  worker_run('z', fn, NULL, 0,0);
}

static void save_flash()
{
  char *fn;
  fn= gfx_select_file(1, NULL, flash_file_types);
  if (!fn) return ;
  worker_run('Z', fn, NULL, 0,0);
}



int main()
{
  int quit;
  char *str;
  char *s;
  char *fn;

  srand(time(NULL));
  gfx_init(NULL,620,700,"WiFi Programmer");
  quit= 0;
  gfx_SCREEN_0();
  gfx_set_log(Wlog);
  show_about();
  read_config();
  if (config.G_flash_file)
    gfx_set_text(Wflashfile, config.flash_file);
  if (config.G_eeprom_file)
    gfx_set_text(Weepromfile, config.eeprom_file);
  main_fd= worker_init();
  if (main_fd<0)
  {
  gfx_logview_printf(Wlog, "Unable to init worker: %s\n",
             strerror(errno));
  }
  else
  {
    gfx_monitor_file(main_fd, 40, 1, 0, 0);
  }
  tick_fd= ticktock_begin(100);
  gfx_monitor_file(tick_fd, 20, 1, 0, 0);
  while(!quit)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    if (tick_event(&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT: quit=1; break;
    case Wconnect:
      if (device_socket!=-1)
      {
        gfx_logview_printf(Wlog, "Already connected, disconnecting.\n");
        close_device(1);
      }
      str= gfx_get_text(Wip);
      if (!str || !str[0]) 
      {
        gfx_logview_printf(Wlog, "Bad address, won't connect.\n");
        break;
      }
      open_device(str);
      worker_run('I', 0, 0, 1, 0);
      break;
    case Wmaint: maintenance_screen(); break;
    case Wselflash:
       fn= gfx_select_file(0, NULL, flash_file_types);
       if (fn) gfx_set_text(Wflashfile, fn);
       break;
    case Wseleeprom:
       fn= gfx_select_file(0, NULL, eeprom_file_types);
       if (fn) gfx_set_text(Weepromfile, fn);
       break;
    case Wsettings: settings_screen(); break;
    case Wpoweron: power_on(); break;
    case Wpoweroff: power_off(0); break;
    case Wrecycle: main_comp_func= continue_recycle; power_off(1); break;
    case Wprgflash: put_prg(1); break;
    case Wprgeeprom: put_prg(2); break;
    case Wprgall: put_prg(3); break;
    case Wfuses: 
       main_comp_func= show_fuses;
       worker_run('u',NULL,NULL,0,0);
       break;
    case Wsaveeeprom: save_eeprom(); break;
    case Wsaveflash: save_flash(); break;
    case Wabout: show_about(); break;
#if 0
  case Wtest:
  worker_run('T', NULL, NULL, strtol(s, NULL, 0),0);
  break;
#endif
    }
  }
  ticktock_end();
  write_config();
  return 0;
}
