
static uint8_t Pm328_speed;

void Pm328_enter_programming_mode()
{
  // setting VCC pin removes power from the target chip
  OUTPUT(PM328_VCC);       SET(PM328_VCC);
  OUTPUT(PM328_MOSI);      CLR(PM328_MOSI);
  OUTPUT(PM328_SCK);       CLR(PM328_SCK);
  INPUT(PM328_MISO);       CLR(PM328_MISO);
  OUTPUT(PM328_RESET);     CLR(PM328_RESET);

  _delay_ms(100);

  CLR(PM328_VCC);

  _delay_ms(100);
}

void Pm328_leave_programming_mode()
{
  SET(PM328_VCC);
  _delay_ms(100);
  CLR(PM328_SCK);          INPUT(PM328_SCK);
  CLR(PM328_MOSI);         INPUT(PM328_MOSI);
                           INPUT(PM328_RESET);
  _delay_ms(5);
  CLR(PM328_VCC);          INPUT(PM328_VCC);
}

static void Pm328_delay()
{
  switch(Pm328_speed)
  {
  case 0: _delay_us(5); break;
  case 1: _delay_us(10); break;
  case 2: _delay_us(25); break;
  case 3: _delay_us(100); break;
  default: _delay_us(500); break;
  }
}

uint8_t Pm328_send_byte(uint8_t V)
{
  uint8_t R;
  int i;
  R= 0;
  for(i=0;i<8;i++)
  {
    if (V&(1<<7)) { SET(PM328_MOSI); }
             else { CLR(PM328_MOSI); }
    Pm328_delay();
    SET(PM328_SCK);
    Pm328_delay();

    R<<= 1;
    if (TST(PM328_MISO)) R|= 1;
    CLR(PM328_SCK);
    V<<= 1;
  }
  return R;
}

uint8_t Pm328_send_command(uint8_t *cmd, int retb)
{
  int i;
  uint8_t R,K;
  R= 0;
  for(i=0;i<4;i++)
  {
    K= Pm328_send_byte(cmd[i]);
    if (i==retb) R= K;
  }
  return R;
}

void Pm328_load_flash_buffer(uint16_t address, uint8_t *data, uint16_t len)
{
  int i;
  uint8_t hilow;

  if (address&1) hilow= 0x08;
            else hilow= 0x00;
 
  for(i=0;i<len;i++)
  {
    Pm328_send_byte( 0x40  | hilow );
    Pm328_send_byte( address >> 9 );
    Pm328_send_byte( address >> 1 );
    Pm328_send_byte( data[i] );
//    _delay_ms(5);
    address++;
    hilow ^= 0x08;
  }
}

void Pm328_read_flash(uint16_t address,uint8_t *data,uint16_t len)
{
  int i;
  uint8_t hilow;

  if (address&1) hilow= 0x08;
            else hilow= 0x00;
 
  for(i=0;i<len;i++)
  {
    Pm328_send_byte( 0x20 | hilow );
    Pm328_send_byte( address >> 9 );
    Pm328_send_byte( address >> 1 );
    data[i]= Pm328_send_byte( 0x00 );
    address++;
    hilow ^= 0x08;
  }
}

void Pm328_recycle_vcc()
{
  // setting VCC pin removes power from the target chip
  OUTPUT(PM328_VCC);       SET(PM328_VCC);
  _delay_ms(500);
  CLR(PM328_VCC);          INPUT(PM328_VCC);
}


