

#define _LCD_E     0x04
#define _LCD_RS    0x80

static void lcd_byte(uint8_t a)
{
  int8_t i;
  for(i=7;i>=0;i--)
  {
    if ((a>>i)&1) { SET(LCD_DATA); } else { CLR(LCD_DATA); }
    SET(LCD_CLOCK);
    _delay_us(1);
    CLR(LCD_CLOCK);
  }
}

static void lcd_rclock()
{
   SET(LCD_RCLOCK);
   _delay_us(1);
   CLR(LCD_RCLOCK);
}

static void lcd_command(uint8_t rs, uint8_t data)
{
  
  lcd_byte(data); lcd_byte(rs | _LCD_E); lcd_rclock();
  _delay_us(1); 
  lcd_byte(data); lcd_byte(rs);          lcd_rclock();
  _delay_us(50); // The document specifies 38us for completion
}

void lcd_clear_display()
{
  lcd_command(0, 1);
                 // Clear Display command
  _delay_ms(2);  // specified as 1.52ms in datasheet
}

void lcd_return_home()
{
  lcd_command(0, 2);
  _delay_ms(2);  // specified as 1.52ms in datasheet
}

void lcd_init()
{
  OUTPUT(LCD_DATA); OUTPUT(LCD_CLOCK); OUTPUT(LCD_RCLOCK);
  _delay_ms(100);  // wait for a while after power up
  lcd_command(0, (1<<5) | (1<<4) | (1<<3));
                 // Function Set command
                          // 8-bit interface
                                   // 2 lines
  lcd_command(0, (1<<3) | (1<<2) );
                 // Display On/Off command
                          // Display On

  lcd_clear_display();
                          
  lcd_command(0, (1<<2) | (1<<1) );
                 // Entry Mode command
                          // Direction= right
}

void lcd_print_char(uint8_t chr)
{
  lcd_command(_LCD_RS, chr);
}

void lcd_set_address(uint8_t row, uint8_t col)
{
  uint8_t addr= row*0x40 + col;
  lcd_command(0,addr|(1<<7));
}


